package com.kusalic.zavrsni;

import java.awt.BorderLayout;
import java.awt.Container;
import java.io.File;
import java.io.IOException;

import javax.swing.JFrame;

/**
 * Klasa ostvaruje eliminacijski genetski algoritam
 * @author Domagoj
 */
public class GA {

	int velPop;
	double vjMut;
	Tripl[] populacija;

	public GA(int velPop,double vjMut) {
		this.velPop=velPop;
		this.vjMut=vjMut;
	}
	
	/**
	 * Inicijalizira populaciju
	 * @return najbolja jedinka
	 * @throws IOException 
	 */
	public IJedinka init(File f) throws IOException {
		ZapisProblema.stvori(f); // Zapisuje problem u ZapisProblema
		
		populacija = new Tripl[velPop];
		for (int i=0;i<velPop;i++)
			populacija[i]=new Tripl(getNextID(),newIJedinka());
		
		return getBest();
	}

	public IJedinka step() {
		int najbolji=0,najgori=0;
		for (int i=1;i<velPop;i++) {
			if (populacija[i].dobrota < populacija[najbolji].dobrota)
				najbolji=i;
			if (populacija[i].dobrota > populacija[najgori].dobrota)
				najgori=i;
		}

		double najboljiD=populacija[najbolji].dobrota; // bolje je manje!
		double najgoriD=populacija[najgori].dobrota;
		int M=(int) (velPop/(2+ ((double)velPop-30)/(100-30)*(4-2)));
		double[] vjElim = new double[velPop];
		double suma=0;
		for (int i=0;i<velPop;i++) {
			vjElim[i]=(populacija[i].dobrota-najboljiD) / (najgoriD-najboljiD);
			suma+=vjElim[i];
		}
		if (suma==0) {
			System.err.println("Populacija je postala monotona!");
			return getBest();
		}

		for (int i=0;i<M*(1+M/velPop);i++) { // eliminiranje otprilike M jedinki iz populacije
			double odabrani=Math.random()*suma;
			for (int j=0;j<velPop;j++) {
				odabrani-=vjElim[j];
				if (odabrani<=1e-7) {
					populacija[j]=null;
					break;
				}
			}
		}
		
		for (int i=0;i<velPop;i++) // nadopuna populacije novim jedinkama
			if (populacija[i]==null) {
				populacija[i]=new Tripl(getNextID(), krizaj());
				if (Math.random()<vjMut) {
					populacija[i].jedinka.mutiraj(); // mutiraj
					populacija[i].dobrota=populacija[i].jedinka.dobrota(); // update dobrote
				}
			}
		
		return getBest();
	}

	IJedinka krizaj() { // krizanje, slucajna selekcija
		int prvi,drugi;
 		do { // prvi roditelj
			prvi=(int) (Math.random()*velPop);
		} while (populacija[prvi]==null);
		do { // drugi roditelj
			drugi=(int) (Math.random()*velPop);
		} while (prvi!=drugi && populacija[drugi]==null);
		// preko bilo koje jedinke se pozove krizaj(a,b) koji stvori novu jedinku
		return populacija[prvi].jedinka.krizaj(
				populacija[prvi].jedinka, populacija[drugi].jedinka); 
	}
	
	/**
	 * @return referencu na najbolju jedinku
	 */
	public IJedinka getBest() {
		int tko=0;
		for (int i=1;i<velPop;i++)
			if (populacija[i].dobrota < populacija[tko].dobrota)
				tko=i;
		return populacija[tko].jedinka;
	}

	public IJedinka[] getAll() {
		IJedinka[] svi=new IJedinka[velPop];
		for (int i=0;i<velPop;i++)
			svi[i]=populacija[i].jedinka;
		return svi;
	}
	
	private static int nextID=0;
	public static int getNextID() {
		nextID++;
		if (nextID>1000000) nextID=0;
		return nextID;
	}
	
	/**
	 * Funkcija obrazca tvornice za lokalizirano stvaranje instanci implementiranog interfacea
	 * @return IJedinka
	 */
	private IJedinka newIJedinka() {
		return new Krojenje();
	}
	
	public static void main(String[] args) throws IOException {
	      JFrame frame=new JFrame();
	        Container cp = frame.getContentPane();
	        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE); 
	        frame.setSize(480,320);
	        cp.setLayout(new BorderLayout());

	        double avg;
	        GA ga = new GA(30,0.3);
	        ga.init(new File("objekti.in"));
	        for (int k=0;k<10;k++) {
	        	for (int i=0;i<10;i++)
	        		System.out.print((int)ga.getAll()[i].dobrota()+" ");  System.out.println();
	        	avg=0; for (int i=0;i<10;i++) avg+=ga.getAll()[i].dobrota();
	        	System.out.println((int)avg/10);
	        	ga.step();
	        }
	        final Prikaz prikaz=new Prikaz((Krojenje)ga.getBest());
	        cp.add(prikaz);
	        frame.setVisible(true);
	}

}

/**
 * Pomocna klasa za pamcenje populacije
 * @author Domagoj
 */
class Tripl {
	int ID;
	IJedinka jedinka;
	double dobrota;

	public Tripl(int iD, IJedinka jedinka) {
		ID = iD;
		this.jedinka = jedinka;
		this.dobrota = jedinka.dobrota();
	}
}

