package com.kusalic.zavrsni;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.lang.reflect.InvocationTargetException;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

/**
 * Klasa omogucuje graficko sucelje za pokretanje genetskog algoritma,
 * te prikaz izvodjenja. <br>
 * @author Domagoj Kusalić
 */
public class GUI extends JFrame {
	private static final long serialVersionUID = 1L;

	private GUI () {
	    super();
	    initGUI();
	}
	
	Container cp;
	JProgressBar bar;
	JLabel status;
	File fajl;
	GA ga;
	JButton radi;
	JButton inicijalizacija;
	Prikaz prikaz;
	
	private void initGUI() {
	    try{
	        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	      } catch (Exception e) {}
	      cp=getContentPane();
	      setBackground(Color.WHITE);
	      setLocation(200,100);
	      setSize(800,640);
	      setTitle("GA - Krojenje");
	      setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE); 
	      cp.setLayout(new BorderLayout());

	      JPanel lijeviPanel = new JPanel();
	      GridLayout gl = new GridLayout(10,1);
	      gl.setVgap(5);
	      lijeviPanel.setLayout(gl);
	      
	      JPanel dodani=new JPanel();
	      dodani.setLayout(new BorderLayout());
	      dodani.add(lijeviPanel);
	      JLabel odvoji1=new JLabel();
	      odvoji1.setPreferredSize(new Dimension(8,1));
	      JLabel odvoji2=new JLabel();
	      odvoji2.setPreferredSize(new Dimension(5,1));
	      dodani.add(BorderLayout.EAST,odvoji1);
	      dodani.add(BorderLayout.WEST,odvoji2);
	      cp.add(BorderLayout.WEST, dodani);
	      bar = new JProgressBar(0, 1000);
	      cp.add(BorderLayout.SOUTH, bar);
	      status=new JLabel("Odaberite datoteku...",JLabel.CENTER);
	      status.setForeground(Color.BLUE);
	      status.setFont(new Font("Serif", Font.BOLD, 16));
	      status.setPreferredSize(new Dimension(200,1));
	      lijeviPanel.add(status);
	      JButton datoteka=new JButton("Učitavanje datoteke");
	      datoteka.setFont(new Font("Times", Font.PLAIN, 16));
	      lijeviPanel.add(datoteka);
	      datoteka.addActionListener(new ActionListener() {
	    	  public void actionPerformed(ActionEvent e) {
	    		  JFileChooser c=new JFileChooser();
	    		    if (c.showOpenDialog(null)==JFileChooser.CANCEL_OPTION) // open
	    		      status.setText("Niste učitali datoteku!");
	    		    else {
	    		      fajl=c.getSelectedFile();
	    		      status.setText("Učitali ste datoteku.");
	    		    };
	    	  }
	      });

	      JPanel mal = new JPanel();
	      mal.setLayout(new GridLayout(1,2));
	      JLabel velPopLabel = new JLabel("<html>Veličina<br>populacije:");
	      velPopLabel.setFont(new Font("Serif", Font.PLAIN, 17));
	      mal.add(velPopLabel);
	      JLabel vjMutLabel = new JLabel("<html>Vjerojatnost<br>mutacije:");
	      vjMutLabel.setFont(new Font("Serif", Font.PLAIN, 17));
	      mal.add(vjMutLabel);
	      lijeviPanel.add(mal);
	      
	      JPanel mali = new JPanel();
	      mali.setLayout(new GridLayout(1,2));
	      final JTextField velPop=new JTextField("30");
	      velPop.setFont(new Font("Serif", Font.PLAIN, 20));
	      mali.add(velPop);
	      final JTextField vjMut=new JTextField("0.3");
	      vjMut.setFont(new Font("Serif", Font.PLAIN, 20));
	      mali.add(vjMut);
	      lijeviPanel.add(mali);
	      
	      inicijalizacija=new JButton("Inicijalizacija");
	      inicijalizacija.setFont(new Font("Times", Font.PLAIN, 16));
	      lijeviPanel.add(inicijalizacija);
	      inicijalizacija.addActionListener(new ActionListener() {
	    	  public void actionPerformed(ActionEvent e) {
	    		  int velPopInt;
	    		  double vjMutBr;
	    		  try {
	    			  velPopInt=Integer.parseInt(velPop.getText());
	    			  vjMutBr=Double.parseDouble(vjMut.getText());
	    		  } catch (NumberFormatException ex) {
	    			  status.setText("Brojevi su krivo upisani.");
	    			  return;
	    		  }
	    		  try {
	    			  ga=new GA(velPopInt,vjMutBr);
	    			  ga.init(fajl);
	    		  } catch (Exception ex) {
	    			  status.setText("Inicijalizacija nije uspjela.");
	    			  return;
	    		  }
	    		  status.setText("Algoritam je inicijaliziran.");
	    		  if (inicijalizacija.getText()!="Restart") {
	    			  inicijalizacija.setText("Restart");
	    			  radi.setEnabled(true);
	    			  prikaz=new Prikaz((Krojenje)ga.getAll()[0]);
	    			  cp.add(prikaz);
	    		  } else {
	    			  prikaz.postavi((Krojenje)ga.getAll()[0]);
	    		  }
	    	  }
	      });
	      
	      JLabel brKor= new JLabel("Koliko koraka želite napraviti:");
	      brKor.setFont(new Font("Serif", Font.PLAIN, 17));
	      lijeviPanel.add(brKor);
	      
	      final JTextField brKoraka=new JTextField("50");
	      brKoraka.setFont(new Font("Serif", Font.PLAIN, 20));
	      lijeviPanel.add(brKoraka);
	      
	      radi=new JButton("Izvrši");
		  radi.setEnabled(false);
	      radi.setFont(new Font("Times", Font.PLAIN, 16));
	      lijeviPanel.add(radi);
	      radi.addActionListener(new ActionListener() {
	    	  public void actionPerformed(ActionEvent e) {
	    		  final int brojIteracija;
	    		  try {
	    			  brojIteracija=Integer.parseInt(brKoraka.getText());
	    		  } catch (NumberFormatException ex) {
	    			  status.setText("Broj je krivo upisan.");
	    			  return;
	    		  }
	    		  inicijalizacija.setEnabled(false);
	    		  radi.setEnabled(false);
	    		  status.setText("Algoritam se izvršava.");
	    		  try {
	    			  new Thread( new Runnable() {
	    				  @Override
	    				  public void run() {
	    					  radi(brojIteracija);
	    				  }
    		          	}).start();
	    		  } catch (Exception ex) {
	    			  status.setText("Graška prilikom izvršavanja.");
	    			  ex.printStackTrace();
	    			  return;
	    		  }
	    	  }
	      });	      
	      setVisible(true);
	}

	private void radi(final int brojKoraka) {
		for (int i=0;i<brojKoraka;i++) {
			ga.step();
			final int ii=i;
			SwingUtilities.invokeLater(new Runnable() {
				public void run() {
					try {
						bar.setValue((ii+1)*1000/brojKoraka);
						prikaz.postavi((Krojenje)ga.getBest());
					} catch (Exception e) {}
		         	}
		       	});
		}
		
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				try {
					prikaz.postavi((Krojenje)ga.getBest());
					status.setText("Algoritam je završio.");
					radi.setEnabled(true);
					inicijalizacija.setEnabled(true);
				} catch (Exception e) {}
	         	}
	       	});
	}
	
	public static void main(String[] args) throws Exception {
	    try { SwingUtilities.invokeAndWait(new Runnable() {
	        public void run() { new GUI(); }
	      }); }
	    catch(InterruptedException e) { e.printStackTrace(); }
	    catch (InvocationTargetException e) { e.printStackTrace(); }
	}		
}
