package com.kusalic.zavrsni;

import java.util.ArrayList;


/**
 * Implementacija sučelja IGenetski za potrebe rješavanja problema krojenja
 * @author Domagoj
 */
public class Krojenje implements IJedinka {

	private ZapisProblema problem;
	private double[] X, Y, kut;
	private int n;

	/**
	 * Stvara jedinku i slučajno je popunjava
	 */
	public Krojenje() {
		problem = ZapisProblema.dajJedinku();
		if (problem == null) {
			System.err.println("Problem nije ucitan!");
			System.exit(1);
		} 
		n = problem.oblici.size();
		X = new double[n];
		Y = new double[n];
		kut = new double[n];
		for (int i = 0; i < n; i++) {
			kut[i] = Math.random() * Math.PI * 2;
			X[i] = Math.random() * 100; // proizvoljna velicina, ovisi o
										// problemu
			Y[i] = Math.random() * 100;
		}
		slijepi();
	}

	@Override
	public double dobrota() {
		double maxX=Integer.MIN_VALUE,minX=Integer.MAX_VALUE,
	       maxY=Integer.MIN_VALUE,minY=Integer.MAX_VALUE;
		for (int i=0;i<n;i++) {
			double tx,ty,x,y;
			ArrayList<Integer> ai=problem.oblici.get(i);
			for (Integer j: ai) {
				tx = problem.Tx.get(j.intValue());
				ty = problem.Ty.get(j.intValue());
				x=tx*Math.cos(kut[i]) - ty*Math.sin(kut[i])  + X[i];
				y=tx*Math.sin(kut[i]) + ty*Math.cos(kut[i])  + Y[i];
				maxX=Math.max(maxX,x);
				maxY=Math.max(maxY,y);
				minX=Math.min(minX,x);
				minY=Math.min(minY,y);
			}
		}
		return (maxX-minX)*(maxY-minY);
	}

	@Override
	public IJedinka krizaj(IJedinka a, IJedinka b) {
		Krojenje k= new Krojenje();
		Krojenje aa=(Krojenje) a;
		Krojenje bb=(Krojenje) b;
		for (int i=0;i<n;i++) { // uniformno krizanje
			if (Math.random()<0.5) {
				k.X[i]=aa.X[i];
				k.Y[i]=aa.Y[i];
				k.kut[i]=aa.kut[i];
			} else {
				k.X[i]=bb.X[i];
				k.Y[i]=bb.Y[i];
				k.kut[i]=bb.kut[i];
			}
		}
		k.slijepi();
		return k;
	}

	@Override
	public void mutiraj() {
		double maxX=Integer.MIN_VALUE,minX=Integer.MAX_VALUE,
	       maxY=Integer.MIN_VALUE,minY=Integer.MAX_VALUE;
		for (int i=0;i<n;i++) {
			double tx,ty,x,y;
			ArrayList<Integer> ai=problem.oblici.get(i);
			for (Integer j: ai) {
				tx = problem.Tx.get(j.intValue());
				ty = problem.Ty.get(j.intValue());
				x=tx*Math.cos(kut[i]) - ty*Math.sin(kut[i])  + X[i];
				y=tx*Math.sin(kut[i]) + ty*Math.cos(kut[i])  + Y[i];
				maxX=Math.max(maxX,x);
				maxY=Math.max(maxY,y);
				minX=Math.min(minX,x);
				minY=Math.min(minY,y);
			}
		}

		int koga=(int)Math.random()*n;
		if (Math.random()<0.5) { // premjestaj
			X[koga]= Math.random()*(maxX-minX) + minX;
			Y[koga]= Math.random()*(maxY-minY) + minY;
		} else // rotacija
			kut[koga]=Math.random() * Math.PI * 2;

		slijepi(); // bitno!
	}

	/**
	 * Vrlo bitna funkcija koja udaljene poligone priblizi koliko god moze (od
	 * blizih prema daljima), radeci binarnu pretragu po udaljenosti od tezista
	 */
	private void slijepi() {
		double[] srednjiX = new double[n];
		double[] srednjiY = new double[n];

		for (int i = 0; i < n; i++) {
			double x, y, tx, ty;
			ArrayList<Integer> ai = problem.oblici.get(i);
			double sumX = 0, sumY = 0;
			for (Integer j : ai) {
				tx = problem.Tx.get(j.intValue());
				ty = problem.Ty.get(j.intValue());
				x = tx * Math.cos(kut[i]) - ty * Math.sin(kut[i])  + X[i];
				y = tx * Math.sin(kut[i]) + ty * Math.cos(kut[i])  + Y[i];
				sumX += x;
				sumY += y;
			}
			srednjiX[i] = sumX / ai.size();
			srednjiY[i] = sumY / ai.size();
		}
		double centarX = 0, centarY = 0;
		for (int i = 0; i < n; i++) {
			centarX += srednjiX[i];
			centarY += srednjiY[i];
		}

		centarX /= n;
		centarY /= n;
		int[] redObrade = new int[n];
		for (int i = 0; i < n; i++)
			redObrade[i] = i;
		for (int i = 0; i < n; i++)
			// poredak po udaljenosti od centra
			for (int j = 0; j < i; j++)
				if ((srednjiX[redObrade[i]] - centarX)
						* (srednjiX[redObrade[i]] - centarX)
						+ (srednjiY[redObrade[i]] - centarY)
						* (srednjiY[redObrade[i]] - centarY) < (srednjiX[redObrade[j]] - centarX)
						* (srednjiX[redObrade[j]] - centarX)
						+ (srednjiY[redObrade[j]] - centarY)
						* (srednjiY[redObrade[j]] - centarY)) {
					redObrade[i] ^= redObrade[j];
					redObrade[j] ^= redObrade[i];
					redObrade[i] ^= redObrade[j];
				}
		// System.out.println(Arrays.toString(redObrade));
		double[] X2 = new double[n]; // nove lokacije poligona
		double[] Y2 = new double[n];
		X2[redObrade[0]] = centarX-(srednjiX[redObrade[0]]-X[redObrade[0]]);
		Y2[redObrade[0]] = centarY-(srednjiY[redObrade[0]]-Y[redObrade[0]]); // postavljamo najblizeg u centar
		for (int i = 1; i < n; i++) {
			double zeljeniCentarX=centarX-(srednjiX[redObrade[i]]-X[redObrade[i]]);
			double zeljeniCentarY=centarY-(srednjiY[redObrade[i]]-Y[redObrade[i]]);
			double deltaX=X[redObrade[i]]-zeljeniCentarX;
			double deltaY=Y[redObrade[i]]-zeljeniCentarY;

			if (Math.abs(deltaX)<1e-3 && Math.abs(deltaY)<1e-3) { deltaX=Math.random()-0.5;  deltaY=Math.random()-0.5; }
			// binarnom pretragom trazimo najmanji faktor s kojim
			// mozemo mnoziti deltaXiY, a da se niste ne sijece
			double maxFaktor=1000000,minFaktor=0,midFaktor;
			while (maxFaktor-minFaktor>0.0001) {
				midFaktor=(maxFaktor+minFaktor)/2;
				if (provjeraZaBinPr(redObrade,X2,Y2,i,
						zeljeniCentarX+deltaX*midFaktor,zeljeniCentarY+deltaY*midFaktor))
					maxFaktor=midFaktor;
				else
					minFaktor=midFaktor;
			}
			X2[redObrade[i]]= zeljeniCentarX+deltaX*maxFaktor; // spremamo najblizu pronadjenu lokaciju
			Y2[redObrade[i]]= zeljeniCentarY+deltaY*maxFaktor;
		}
		for (int i=0;i<n;i++) {
			X[i]=X2[i]; Y[i]=Y2[i]; }
	}

	/**
	 * Vraca true ukoliko se poligoni ne sijecu i nisu ugnjezdjeni. <br>
	 * Koristi ga binarna pretraga.
	 * @return true ukoliko je dodani poligon van ostalih
	 */
	private boolean provjeraZaBinPr(int[] redObrade, double[] X2, double[] Y2,
			int N, double xPos, double yPos) {
		// provjera sjece li se koji brid:
		ArrayList<Integer> trenutniPoli = problem.oblici.get(redObrade[N]);
		double[] poliX = new double[trenutniPoli.size()+1]; // u krug idu indeksi (ima dodatni)
		double[] poliY = new double[trenutniPoli.size()+1];
		for (int i=0;i<trenutniPoli.size();i++) {
			double tx=problem.Tx.get(trenutniPoli.get(i).intValue());
			double ty=problem.Ty.get(trenutniPoli.get(i).intValue());
			poliX[i] = tx * Math.cos(kut[redObrade[N]]) - ty * Math.sin(kut[redObrade[N]]) + xPos;
			poliY[i] = tx * Math.sin(kut[redObrade[N]]) + ty * Math.cos(kut[redObrade[N]]) + yPos;
		}
		poliX[poliX.length-1]=poliX[0]; // zatvaranje kruga
		poliY[poliY.length-1]=poliY[0];
		
		double[] zaProvjeritiX = new double[N];
		double[] zaProvjeritiY = new double[N];
		int provjeraUnutra1=0,provjeraUnutra2=0;
		for (int ii=0;ii<N;ii++) {
			int i = redObrade[ii]; // baratam po blizini, ne po indeksu, te sa X2, ne X
			ArrayList<Integer> ai = problem.oblici.get(i);
			double Lx, Ly, x, y, tx, ty;
			tx = problem.Tx.get(ai.get(ai.size()-1).intValue());
			ty = problem.Ty.get(ai.get(ai.size()-1).intValue());
			Lx = tx * Math.cos(kut[i]) - ty * Math.sin(kut[i]) + X2[i];
			Ly = tx * Math.sin(kut[i]) + ty * Math.cos(kut[i]) + Y2[i];
			zaProvjeritiX[ii]=Lx; // da bi provjerili je li poligon obujmio neki stari
			zaProvjeritiY[ii]=Ly;
			for (Integer j : ai) {
				tx = problem.Tx.get(j.intValue());
				ty = problem.Ty.get(j.intValue());
				x = tx * Math.cos(kut[i]) - ty * Math.sin(kut[i]) + X2[i];
				y = tx * Math.sin(kut[i]) + ty * Math.cos(kut[i]) + Y2[i];

				for (int k=0;k<poliX.length-1;k++) // provjera!
					if (sijece(Lx,Ly,x,y,poliX[k],poliY[k],poliX[k+1],poliY[k+1]))
						return false;
				if (((CCW(poliX[0],poliY[0],1e6,0,Lx,Ly)<1 && // je li u poligonu?
						  CCW(poliX[0],poliY[0],1e6,0,x,y) == 1) ||
						 (CCW(poliX[0],poliY[0],1e6,0,Lx,Ly)>-1 &&
	  					  CCW(poliX[0],poliY[0],1e6,0,x,y) == -1))  &&
	  					  CCW(Lx,Ly,x,y,poliX[0],poliY[0])*CCW(Lx,Ly,x,y,1e6,0)<1 )
						provjeraUnutra1++;
				if (((CCW(poliX[0],poliY[0],0,1e6,Lx,Ly)<1 && // dvaput!
						  CCW(poliX[0],poliY[0],0,1e6,x,y) == 1) ||
						 (CCW(poliX[0],poliY[0],0,1e6,Lx,Ly)>-1 &&
	  					  CCW(poliX[0],poliY[0],0,1e6,x,y) == -1))  &&
	  					  CCW(Lx,Ly,x,y,poliX[0],poliY[0])*CCW(Lx,Ly,x,y,0,1e6)<1 )
						provjeraUnutra2++;
				
				Lx = x;
				Ly = y;
			}
		}
		if (provjeraUnutra1%2==1 || provjeraUnutra2%2==1)
			return false; 
		
		for (int j=0;j<N;j++) { // provjera obujmivanja!
			int sjecista=0;
			for (int k=0;k<poliX.length-1;k++) {
				if (((CCW(zaProvjeritiX[j],zaProvjeritiY[j],1e6,0,poliX[k],poliY[k])<1 &&
					  CCW(zaProvjeritiX[j],zaProvjeritiY[j],1e6,0,poliX[k+1],poliY[k+1]) == 1) ||
					 (CCW(zaProvjeritiX[j],zaProvjeritiY[j],1e6,0,poliX[k],poliY[k])>-1 &&
					  CCW(zaProvjeritiX[j],zaProvjeritiY[j],1e6,0,poliX[k+1],poliY[k+1]) == -1))  &&
	  				  CCW(poliX[k],poliY[k],poliX[k+1],poliY[k+1],zaProvjeritiX[j],zaProvjeritiY[j])
	  				 *CCW(poliX[k],poliY[k],poliX[k+1],poliY[k+1],1e6,0)<1 )
						sjecista++;
			}
			if (sjecista%2==1) // obujmio je
				return false;
		}
		
		return true;
	}

	/**
	 * sjecu li se (ili dodiruju) duzina A-B i duzina C-D?
	 */
	private boolean sijece(double Ax, double Ay, double Bx, double By,
						   double Cx, double Cy, double Dx, double Dy) { 
//		System.out.println("sijeku "+(int)Ax+","+(int)Ay+" "+(int)Bx+","+(int)By+" "+(int)Cx+","+(int)Cy+" "+(int)Dx+","+(int)Dy+" "+
//				(CCW(Ax,Ay,Bx,By,Cx,Cy)*CCW(Ax,Ay,Bx,By,Dx,Dy)!=1 &&
//				   CCW(Cx,Cy,Dx,Dy,Ax,Ay)*CCW(Cx,Cy,Dx,Dy,Bx,By)!=1)
//		);
		
		return CCW(Ax,Ay,Bx,By,Cx,Cy)*CCW(Ax,Ay,Bx,By,Dx,Dy)<1 &&
			   CCW(Cx,Cy,Dx,Dy,Ax,Ay)*CCW(Cx,Cy,Dx,Dy,Bx,By)<1;
	}
	
	private int CCW(double Ax,double Ay, double Bx,double By,double Cx,double Cy) {
		double value = Ax*By-Ay*Bx + Bx*Cy-By*Cx + Cx*Ay-Cy*Ax;
		double eps = Math.max(Math.max(Math.abs(Ax),Math.abs(Ay)),Math.max(
							  Math.max(Math.abs(Bx),Math.abs(By)),
							  Math.max(Math.abs(Cx),Math.abs(Cy))))/1e10;
		if (value <= -eps) return -1;
		if (value >= eps) return 1;
		return 0;
	}

	public double[] getX() {
		return X;
	}

	public double[] getY() {
		return Y;
	}

	public double[] getKut() {
		return kut;
	}

	public ZapisProblema getProblem() {
		return problem;
	}

}
